/* $Id: ipcif.c,v 1.1.1.1 2004/05/09 03:04:09 tkudou Exp $ */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/sem.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <assert.h>

#define WITH_IPCIF
#include "ipcif.h"

#ifdef _SEM_SEMUN_UNDEFINED
union semun
{
  int val;			/* value for SETVAL */
  struct semid_ds *buf;		/* buffer for IPC_STAT & IPC_SET */
  unsigned short int *array;	/* array for GETALL & SETALL */
  struct seminfo *__buf;	/* buffer for IPC_INFO */
};
#endif /* _SEM_SEMUN_UNDEFINED */

static void ipcif_die(char* msg);
static void ipcif_die_with_errno(int err_no, char* msg);

static void ipcif_initializeMsgq();
static void ipcif_finalizeMsgq(IPCIF_Bool remove_resource_p);
static void ipcif_clearShmScreen();
static void ipcif_initializeShm();
static void ipcif_finalizeShm(IPCIF_Bool remove_resource_p);
static void ipcif_initializeSemaphore();
static void ipcif_finalizeSemaphore(IPCIF_Bool remove_resource_p);

#ifndef MSGMAX
#define MSGMAX 8192
#endif /* MSGMAX */

/* the id of the message queue */
static int ipcif_qid = 0;

/* the id of the shared memory */
static int ipcif_shmid = 0;

/* the attached shared memory */
static IPCIF_Shm* ipcif_shm = 0;

/* the semaphore */
static int ipcif_semid = 0;

static void
ipcif_die(char* msg)
{
  fprintf(stderr, "IPCIF FATAL ERROR: ");
  fprintf(stderr, msg);
  fflush(stderr);
  exit(1);
}

static void
ipcif_die_with_errno(int err_no, char* msg)
{
  fprintf(stderr, "IPCIF FATAL ERROR: ");
  fprintf(stderr, "errno: %d ", err_no);
  fprintf(stderr, msg);
  fprintf(stderr, "\n");
  fflush(stderr);
  exit(1);
}

/*!
 * This function initialize the message queue.
 */
static void
ipcif_initializeMsgq()
{
  key_t cmd_key = IPCIF_MSG_KEY;

  ipcif_qid = msgget(cmd_key, 0);
  if (ipcif_qid < 0)
  {
    ipcif_qid = msgget(cmd_key, IPC_CREAT | 0600);
  }
  if (ipcif_qid < 0)
  {
    ipcif_die("can't open ipc message queue identifier.\n");
    /* not reached */
  }
}

static void
ipcif_finalizeMsgq(IPCIF_Bool remove_resource_p)
{
  if (remove_resource_p)
  {
    if (msgctl(ipcif_qid, IPC_RMID, 0) < 0)
    {
      ipcif_die("can't remove the IPC queue.\n");
      /* not reached */
    }
  }
}

/*!
 * This function clear screen
 */
static void
ipcif_clearShmScreen()
{
  int x, y;
  for (y = 0; y < IPCIF_SHM_DIM_Y_MAX; y++)
  {
    for (x = 0; x < IPCIF_SHM_DIM_X_MAX; x++)
    {
      ipcif_shm->img[y][x] = ' ';
      ipcif_shm->attr[y][x] = 0;
    }
  }
}

/*!
 *  This function initialize the shared memory.
 */
static void
ipcif_initializeShm()
{
  IPCIF_Bool created_p = IPCIF_False;
  ipcif_lockSemaphore();
  key_t key = IPCIF_SHM_KEY;
  ipcif_shmid = shmget(key, IPCIF_SHM_SIZE, 0);
  if (ipcif_shmid < 0)
  {
    ipcif_shmid = shmget(key, IPCIF_SHM_SIZE, IPC_CREAT | 0700);
    created_p = IPCIF_True;
  }
  if (ipcif_shmid < 0)
  {
    ipcif_die_with_errno(errno, "shmget");
    /* not reached */
  }
  ipcif_shm = (IPCIF_Shm*)shmat(ipcif_shmid, 0, 0);
  if ((void*)ipcif_shm == (void*)-1)
  {
    ipcif_die_with_errno(errno, "shmat");
    /* not reached */
  }
  if (created_p)
  {
    /* minimal initialisation just to have valid data */
    ipcif_shm->dim_x = 80;
    ipcif_shm->dim_y = 1;
    ipcif_shm->cursor_x = 0;
    ipcif_shm->cursor_y = 0;
    ipcif_clearShmScreen();
    strcpy(&(ipcif_shm->img[0][0]), "screen is initializing...");
  }
  ipcif_unlockSemaphore();
}

static void
ipcif_finalizeShm(IPCIF_Bool remove_resource_p)
{
  if (remove_resource_p)
  {
    if (shmctl(ipcif_shmid, IPC_RMID, 0) == -1)
    {
      ipcif_die("can't remove the shared memory.\n");
      /* not reached */
    }
  }
  else
  {
    /* only detach */
    if (shmdt(ipcif_shm) == -1)
    {
      ipcif_die("can't detach the shared memory.\n");
      /* not reached */
    }
  }
}

IPCIF_Shm*
ipcif_getShm()
{
  return ipcif_shm;
}

static void
ipcif_initializeSemaphore()
{
#ifndef NOT_USE_IPCIF_SEM_CTRL
  key_t sem_key = IPCIF_SEM_KEY;
  union semun arg;

  ipcif_semid = semget(sem_key, 1, IPC_CREAT | IPC_EXCL | 0600);
  if (ipcif_semid == -1)
  {
    switch (errno)
    {
     case EACCES:
      printf("ERRNO: EACCES\n");
      break;
     case EEXIST:
      printf("ERRNO: EEXIST\n");
      break;
     case ENOENT:
      printf("ERRNO: ENOENT\n");
      break;
     case EINVAL:
      printf("ERRNO: EINVAL\n");
      break;
     case ENOMEM:
      printf("ERRNO: ENOMEM\n");
      break;
     case ENOSPC:
      printf("ERRNO: ENOSPC\n");
      break;
     default:
      printf("ERRNO: %d\n", errno);
      break;
    }
    fflush(stdout);
    ipcif_die("can't create semaphore.\n");
  }
  else if (ipcif_semid > 0)
  {
    struct sembuf sb[1];

    /* success, this semaphore have to be initialized. */
    arg.val = 0;
    if (semctl(ipcif_semid, 0, SETVAL, arg) == -1)
    {
      ipcif_die("can't initialize the semaphore.\n");
      /* not reached */
    }

    /* initialize the semaphore value and set the sem_otime. */
    sb[0].sem_num = 0;
    sb[0].sem_op = +1;
    sb[0].sem_flg = 0;
  
    if (semop(ipcif_semid, sb, 1) == -1)
    {
      printf("ERRNO: %d\n", errno);
      fflush(stdout);
      ipcif_die("initialization semop(+1)\n");
    }
    /* printf("The semaphore is created.\n");*/
  }
  else
  {
    const int max_try = 10;
    struct semid_ds ds;
    struct timespec t;
    int n;
    
    t.tv_sec = 1;
    t.tv_nsec = 0;
    arg.buf = &ds;

    /* only open */
    ipcif_semid = semget(sem_key, 1, 0600);
    
    /* This process have to wait for the initialization of this semaphore. */
    for (n = 0; n < max_try; n++)
    {
      if (semctl(ipcif_semid, 0, IPC_STAT, arg) == -1)
      {
	ipcif_die("can't wait the initialization of the semaphore.\n");
      }
      if (arg.buf->sem_otime != 0)
      {
	break;
      }
#if 0
      printf("X");
      fflush(stdout);
#endif
      nanosleep(&t, NULL); /* get a sleep, 1 sec. */
    }
    if (n == max_try)
    {
      ipcif_die("semget OK, but semaphore isn't initialized.\n");
    }
    /* printf("The semaphore is opened.\n"); */
  }
  if (ipcif_semid < 0)
  {
    ipcif_die("can't get ipc semaphore identifier.\n");
    /* not reached */
  }
#endif /* NOT_USE_IPCIF_SEM_CTRL */
}

static void
ipcif_finalizeSemaphore(IPCIF_Bool remove_resource_p)
{
#ifndef NOT_USE_IPCIF_SEM_CTRL
  if (remove_resource_p)
  {
    if (semctl(ipcif_semid, 0, IPC_RMID) == -1)
    {
      ipcif_die("can't remove ipc semaphore.\n");
      /* not reached */
    }
  }
#endif /* NOT_USE_IPCIF_SEM_CTRL */
}

IPCIF_Bool
ipcif_initialize()
{
  ipcif_initializeMsgq();
  ipcif_initializeSemaphore();
  ipcif_initializeShm();
  
  return IPCIF_Success;
}

IPCIF_Bool
ipcif_finalize(IPCIF_Bool remove_resource_p)
{
  ipcif_finalizeShm(remove_resource_p);
  ipcif_finalizeSemaphore(remove_resource_p);
  ipcif_finalizeMsgq(remove_resource_p);
  
  return IPCIF_Success;
}

#define IPCIF_MAX_DATA_LENGTH (MSGMAX - (sizeof(IPCIF_Msg) - sizeof(long)))

int
ipcif_max_data_length()
{
  return IPCIF_MAX_DATA_LENGTH;
}

IPCIF_Msg*
ipcif_initMsg(long cmd_no, IPCIF_Sender sender, int cmd_data_length, char* cmd_data)
{
  IPCIF_Msg* msg;

  if (cmd_data_length > IPCIF_MAX_DATA_LENGTH)
  {
    return 0;
  }
  if (cmd_data_length > 0)
  {
    msg = (IPCIF_Msg*)malloc(sizeof(IPCIF_Msg) + cmd_data_length);
  }
  else
  {
    msg = (IPCIF_Msg*)malloc(sizeof(IPCIF_Msg));
  }
  if (!msg)
  {
    return 0;
  }
  msg->cmd_no = cmd_no;
  msg->sender = sender;
  msg->cmd_data_length = cmd_data_length;
  if (cmd_data != 0 && cmd_data_length > 0)
  {
    memcpy(msg->cmd_data, cmd_data, cmd_data_length);
  }

  return msg;
}

void
ipcif_deleteMsg(IPCIF_Msg* msg)
{
  if (msg)
  {
    free((void*)msg);
  }
}

#define IPCIF_GET_MSG_SIZE(msg) (sizeof(IPCIF_Msg) - sizeof(long) + msg->cmd_data_length)

IPCIF_Bool
ipcif_sendMsg(IPCIF_Msg* msg)
{
  int msg_size;
  
  if (!msg)
  {
    return IPCIF_Failure;
  }
  if (msg->cmd_data_length > IPCIF_MAX_DATA_LENGTH)
  {
    return IPCIF_Failure;
  }
  msg_size = IPCIF_GET_MSG_SIZE(msg);
  if (msgsnd(ipcif_qid, (void*)msg, msg_size, 0) < 0)
  {
    /* can't send this message */
    return IPCIF_Failure;
  }
  return IPCIF_Success;
}

IPCIF_Msg*
ipcif_readMsg()
{
  IPCIF_Msg* msg;

  msg = (IPCIF_Msg*)malloc(sizeof(IPCIF_Msg) + IPCIF_MAX_DATA_LENGTH);
  if (!msg)
  {
    return 0;
  }
  if (msgrcv(ipcif_qid, msg, MSGMAX, 0, IPC_NOWAIT) < 0)
  {
    free((void*)msg);
    return 0;
  }
  return msg;
}

IPCIF_Msg*
ipcif_readMsgWithCmdNo(long cmd_no)
{
  IPCIF_Msg* msg;

  msg = (IPCIF_Msg*)malloc(sizeof(IPCIF_Msg) + IPCIF_MAX_DATA_LENGTH);
  if (!msg)
  {
    return 0;
  }
  if (msgrcv(ipcif_qid, msg, MSGMAX, cmd_no, IPC_NOWAIT) < 0)
  {
    free((void*)msg);
    return 0;
  }
  return msg;
}

void
ipcif_lockSemaphore()
{
#ifndef NOT_USE_IPCIF_SEM_CTRL
  struct sembuf sb[1];
  
  sb[0].sem_num = 0;
  sb[0].sem_op = -1;
  sb[0].sem_flg = 0;
  
  if (semop(ipcif_semid, sb, 1) == -1)
  {
    printf("ERRNO: %d\n", errno);
    fflush(stdout);
    ipcif_die("lock semop(+1)\n");
  }
#endif /* NOT_USE_IPCIF_SEM_CTRL */
}

void
ipcif_unlockSemaphore()
{
#ifndef NOT_USE_IPCIF_SEM_CTRL
  struct sembuf sb[1];
  
  sb[0].sem_num = 0;
  sb[0].sem_op = +1;
  sb[0].sem_flg = 0;
  
  if (semop(ipcif_semid, sb, 1) == -1)
  {
    printf("ERRNO: %d\n", errno);
    fflush(stdout);
    ipcif_die("unlock semop(-1)\n");
  }
#endif /* NOT_USE_IPCIF_SEM_CTRL */
}

IPCIF_Bool
ipcif_hasLockP()
{
#ifndef NOT_USE_IPCIF_SEM_CTRL
  union semun arg;
  int sem_pid;
  int sem_val;
  arg.val = 0;

  if ((sem_val = semctl(ipcif_semid, 0, GETVAL, arg)) == -1)
  {
    ipcif_die("can't get the status of the semaphore. (1)\n");
  }
  if (sem_val > 0)
  {
    /* not locked */
    return IPCIF_False;
  }
  
  if ((sem_pid = semctl(ipcif_semid, 0, GETPID, arg)) == -1)
  {
    ipcif_die("can't get the status of the semaphore. (2)\n");
    /* not reached */
  }
  return (getpid() == sem_pid) ? IPCIF_True : IPCIF_False;
#else /* NOT_USE_IPCIF_SEM_CTRL */
  return IPCIF_False;
#endif /* NOT_USE_IPCIF_SEM_CTRL */
}
